"""
Referencing implementations for JSON Schema specs (historic & current).
"""

from __future__ import annotations

from urllib.parse import urljoin

from referencing._attrs import frozen
from referencing._core import (
    Anchor,
    IdentifiedResource,
    Resolver,
    Specification,
)
from referencing.typing import ObjectSchema, Schema


def _dollar_id(resource: Schema):
    if resource is True or resource is False:
        return
    return resource.get("$id")


def _dollar_id_pre2019(resource: Schema):
    if resource is True or resource is False or "$ref" in resource:
        return None
    id = resource.get("$id")
    if id is not None and not id.startswith("#"):
        return id


def _id_no_ref(resource):
    if "$ref" not in resource:
        id = resource.get("id")
        if id is not None and not id.startswith("#"):
            return id


def _anchor_or_dynamic_anchor(
    resource: ObjectSchema,
    specification: Specification,
):
    anchor = resource.get("$anchor")
    if anchor is not None:
        yield Anchor(
            name=anchor,
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        )

    dynamic_anchor = resource.get("$dynamicAnchor")
    if dynamic_anchor is not None:
        yield DynamicAnchor(
            name=dynamic_anchor,
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        )


def _subresources_of(subresource, in_items, in_values):
    def subresources_of(resource):
        for each in subresource:
            if each in resource:
                yield resource[each]
        for each in in_items:
            if each in resource:
                yield from resource[each]
        for each in in_values:
            if each in resource:
                yield from resource[each].values()

    return subresources_of


def _subresources_of_pre2020(subresource, in_items, in_values):
    def subresources_of(resource):
        for each in subresource:
            if each in resource:
                yield resource[each]
        for each in in_items:
            if each in resource:
                yield from resource[each]
        for each in in_values:
            if each in resource:
                yield from resource[each].values()

        items = resource.get("items")
        if items is None:
            return
        elif isinstance(items, list):
            yield from items
        else:
            yield items

    return subresources_of


DRAFT202012 = Specification(
    id_of=_dollar_id,
    anchors_in=_anchor_or_dynamic_anchor,
    subresources_of=_subresources_of(
        subresource={"items", "not"},
        in_items={"allOf"},
        in_values={"$defs", "properties"},
    ),
)


DRAFT201909 = Specification(
    id_of=_dollar_id,
    anchors_in=lambda resource, specification: [
        Anchor(
            name=resource["$anchor"],
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        ),
    ]
    if "$anchor" in resource
    else [],
    subresources_of=_subresources_of_pre2020(
        subresource={"not"},
        in_items={"allOf"},
        in_values={"$defs", "properties"},
    ),
)


DRAFT7 = Specification(
    id_of=_dollar_id_pre2019,
    anchors_in=lambda resource, specification: [
        Anchor(
            name=resource["$id"][1:],
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        ),
    ]
    if resource.get("$id", "").startswith("#")
    else [],
    subresources_of=_subresources_of_pre2020(
        subresource={"not"},
        in_items={"allOf"},
        in_values={"definitions", "properties"},
    ),
)


DRAFT6 = Specification(
    id_of=_dollar_id_pre2019,
    anchors_in=lambda resource, specification: [
        Anchor(
            name=resource["$id"][1:],
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        ),
    ]
    if resource.get("$id", "").startswith("#")
    else [],
    subresources_of=_subresources_of_pre2020(
        subresource={"not"},
        in_items={"allOf"},
        in_values={"definitions", "properties"},
    ),
)


DRAFT4 = Specification(
    id_of=_id_no_ref,
    anchors_in=lambda resource, specification: [
        Anchor(
            name=resource["id"][1:],
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        ),
    ]
    if resource.get("id", "").startswith("#")
    else [],
    subresources_of=_subresources_of_pre2020(
        subresource={"not"},
        in_items={"allOf"},
        in_values={"definitions", "properties"},
    ),
)


DRAFT3 = Specification(
    id_of=_id_no_ref,
    anchors_in=lambda resource, specification: [
        Anchor(
            name=resource["id"][1:],
            resource=IdentifiedResource(
                resource=resource,
                specification=specification,
            ),
        ),
    ]
    if resource.get("id", "").startswith("#")
    else [],
    subresources_of=_subresources_of_pre2020(
        subresource={"not"},
        in_items={"allOf"},
        in_values={"definitions", "properties"},
    ),
)


@frozen
class DynamicAnchor:
    """
    Dynamic anchors, introduced in draft 2020.
    """

    name: str
    resource: IdentifiedResource

    def resolve(self, resolver, uri):
        last = self.resource
        for _, resource, anchors in resolver.dynamic_scope():
            anchor = anchors.get(self.name)
            if isinstance(anchor, DynamicAnchor):
                last = anchor.resource
            elif "$ref" not in resource.resource:
                break
        return last, last.id() or ""


def lookup_recursive_ref(
    resolver: Resolver,
    recursiveRef: str,
) -> tuple[Schema, Resolver]:
    """
    Recursive references (via recursive anchors), present only in draft 2019.
    """
    subschema, resolver = resolver.lookup(recursiveRef)
    if subschema.get("$recursiveAnchor"):  # type: ignore # FIXME: missing test
        for uri, _, _ in resolver.dynamic_scope():
            ref = urljoin(uri, recursiveRef)
            next_subschema, next_resolver = resolver.lookup(ref)
            if not next_subschema.get("$recursiveAnchor"):  # type: ignore
                break
            subschema, resolver = next_subschema, next_resolver
    return subschema, resolver


BY_ID = {
    "https://json-schema.org/draft/2020-12/schema": DRAFT202012,
    "https://json-schema.org/draft/2019-09/schema": DRAFT201909,
    "http://json-schema.org/draft-07/schema#": DRAFT7,
    "http://json-schema.org/draft-06/schema#": DRAFT6,
    "http://json-schema.org/draft-04/schema#": DRAFT4,
    "http://json-schema.org/draft-03/schema#": DRAFT3,
}



from pathlib import Path
import importlib.metadata
import re

from hyperlink import URL
from sphinx.ext.intersphinx import resolve_reference_in_inventory

DOCS = Path(__file__).parent

GITHUB = URL.from_text("https://github.com/")
HOMEPAGE = GITHUB.child("python-jsonschema", "referencing")

project = "referencing"
author = "Julian Berman"
copyright = "2022, " + author

release = importlib.metadata.version("referencing")
version = release.partition("-")[0]

language = "en"
default_role = "any"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_click",
    "sphinx_copybutton",
    "sphinx_json_schema_spec",
    "sphinxcontrib.spelling",
]

pygments_style = "lovelace"
pygments_dark_style = "one-dark"

html_theme = "furo"
html_static_path = []

# See sphinx-doc/sphinx#10785
_TYPE_ALIASES = dict(
    AnchorType=("class", "Anchor"),
    ObjectSchema=("data", "ObjectSchema"),
    Schema=("data", "Schema"),
)


def _resolve_broken_refs(app, env, node, contnode):
    if node["refdomain"] != "py":
        return

    # Evade tobgu/pyrsistent#267
    if node["reftarget"].startswith("pyrsistent.typing."):
        node["reftarget"] = node["reftarget"].replace(".typing.", ".")
        return resolve_reference_in_inventory(
            env, "pyrsistent", node, contnode
        )
    elif node["reftarget"] == "PList":
        node["reftarget"] = "pyrsistent.PList"
        return resolve_reference_in_inventory(
            env, "pyrsistent", node, contnode
        )

    kind, target = _TYPE_ALIASES.get(node["reftarget"], (None, None))
    if kind is not None:
        return app.env.get_domain("py").resolve_xref(
            env,
            node["refdoc"],
            app.builder,
            kind,
            target,
            node,
            contnode,
        )


def setup(app):
    app.connect("missing-reference", _resolve_broken_refs)


# -- Extension configuration -------------------------------------------------

# -- Options for autodoc extension -------------------------------------------

autodoc_default_options = {
    "members": True,
    "member-order": "bysource",
}

# -- Options for autosectionlabel extension ----------------------------------

autosectionlabel_prefix_document = True

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "pyrsistent": ("https://pyrsistent.readthedocs.io/en/latest/", None),
    "python": ("https://docs.python.org/", None),
}

# -- Options for extlinks extension ------------------------------------------

extlinks = {
    "gh": (str(HOMEPAGE.child("%s")), None),
    "github": (str(GITHUB.child("%s")), None),
}

# -- Options for the linkcheck builder ---------------------------------------


def entire_domain(host):
    return r"http.?://" + re.escape(host) + r"($|/.*)"


linkcheck_ignore = [
    entire_domain("codecov.io"),
    entire_domain("img.shields.io"),
    f"{GITHUB}.*#.*",
    str(HOMEPAGE.child("actions")),
    str(HOMEPAGE.child("python-jsonschema/workflows/CI/badge.svg")),
]

# -- Options for spelling extension ------------------------------------------

spelling_show_suggestions = True



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from attrs import define, frozen
else:
    from attrs import define as _define, frozen as _frozen

    def define(cls):
        cls.__init_subclass__ = UnsupportedSubclassing.complain
        return _define(cls)

    def frozen(cls):
        cls.__init_subclass__ = UnsupportedSubclassing.complain
        return _frozen(cls)


class UnsupportedSubclassing(Exception):
    @classmethod
    def complain(this):
        raise UnsupportedSubclassing(
            "Subclassing is not part of referencing's public API. "
            "If no other suitable API exists for what you're trying to do, "
            "feel free to file an issue asking for one.",
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, Union

try:
    from collections.abc import Mapping

    Mapping[str, str]
except TypeError:
    from typing import Mapping

if TYPE_CHECKING:
    from referencing._core import IdentifiedResource, Resolver


#: A JSON Schema which is a JSON object
ObjectSchema = Mapping[str, Any]

#: A JSON Schema of any kind
Schema = Union[bool, ObjectSchema]


class Anchor(Protocol):
    """
    An anchor within a `Schema`.

    Beyond "simple" anchors, some specifications like JSON Schema's 2020
    version have dynamic anchors.
    """

    @property
    def name(self) -> str:
        ...

    def resolve(
        self,
        resolver: Resolver,
        uri: str,
    ) -> tuple[IdentifiedResource, str]:
        pass



from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from urllib.parse import unquote, urldefrag, urljoin

from attrs import evolve, field
from pyrsistent import m, plist, pmap, s
from pyrsistent.typing import PList, PMap, PSet

from referencing._attrs import define, frozen
from referencing.typing import Anchor as AnchorType, ObjectSchema, Schema


class UnidentifiedResource(Exception):
    pass


@frozen
class IdentifiedResource:

    _specification: Specification
    resource: Schema

    @classmethod
    def from_resource(
        cls,
        resource: Schema,
        default_specification: Specification = ...,  # type: ignore
    ):
        specification = default_specification

        if resource is not True and resource is not False:
            jsonschema_schema_keyword = resource.get("$schema")
            if jsonschema_schema_keyword is not None:
                from referencing import jsonschema

                specification = jsonschema.BY_ID.get(
                    jsonschema_schema_keyword,
                    default_specification,
                )

        if specification is ...:
            raise UnidentifiedResource(resource)
        return cls(resource=resource, specification=specification)

    def pointer(self, base_uri: str, pointer: str) -> tuple[str, Schema]:
        """
        Resolve the given JSON pointer, returning a base URI and resource.
        """
        resource = self.resource
        for segment in unquote(pointer[1:]).split("/"):
            if isinstance(resource, Sequence):
                segment = int(segment)  # type: ignore
            else:
                segment = segment.replace("~1", "/").replace("~0", "~")
            resource = resource[segment]  # type: ignore # this can't be a bool
            # FIXME: this is wrong, we need to know that we are crossing
            #        the boundary of a *schema* specifically
            if not isinstance(resource, Sequence):
                id = self._specification.id_of(resource)
                if id is not None:
                    base_uri = urljoin(base_uri, id)
        return base_uri, resource

    def id(self):
        return self._specification.id_of(self.resource)

    def anchors(self) -> list[AnchorType]:
        if isinstance(self.resource, bool):
            return []
        return self._specification.anchors_in(self.resource)

    def subresources(self) -> Iterable[IdentifiedResource]:
        subresources = self._specification.subresources_of(self.resource)  # type: ignore  # FIXME: missing test  # noqa: E501
        return (
            IdentifiedResource.from_resource(
                resource=each,
                default_specification=self._specification,
            )
            for each in subresources
            if each is not True and each is not False
        )


@frozen
class Anchor:

    name: str
    resource: IdentifiedResource

    def resolve(self, resolver, uri):
        return self.resource, uri


@frozen
class Specification:
    """
    A referencing-defining specification.

    See `referencing.jsonschema` for JSON Schema-specific instances.
    """

    id_of: Callable[[Schema], str | None]
    subresources_of: Callable[[ObjectSchema], Iterable[Schema]]
    _anchors_in: Callable[[ObjectSchema, Specification], list[AnchorType]]

    def anchors_in(self, resource: ObjectSchema) -> list[AnchorType]:
        return self._anchors_in(resource, self)


#: A 'null' specification, where resources are opaque
#: (e.g. have no subresources or IDs).
OPAQUE_SPECIFICATION = Specification(
    id_of=lambda resource: None,
    anchors_in=lambda resource, specification: [],
    subresources_of=lambda resource: (),
)


@frozen
class Registry:

    _contents: PMap[
        str,
        tuple[IdentifiedResource, PMap[str, AnchorType]],
    ] = field(
        default=m(),
        repr=lambda value: f"({len(value)} entries)",
    )
    _uncrawled: PSet[str] = field(default=s(), repr=False)

    def update(self, *registries: Registry) -> Registry:
        contents = (each._contents for each in registries)
        uncrawled = (each._uncrawled for each in registries)
        return evolve(
            self,
            contents=self._contents.update(*contents),
            uncrawled=self._uncrawled.update(*uncrawled),
        )

    def with_resource(self, resource: Schema) -> Registry:
        identified = IdentifiedResource.from_resource(resource)
        return self.with_identified_resource(
            uri=identified.id(),
            resource=identified,
        )

    def with_resources(
        self,
        pairs: Iterable[tuple[str, Schema]],
        **kwargs,
    ) -> Registry:
        return self.with_identified_resources(
            (uri, IdentifiedResource.from_resource(resource, **kwargs))
            for uri, resource in pairs
        )

    def with_identified_resource(
        self,
        uri: str,
        resource: IdentifiedResource,
    ) -> Registry:
        return self.with_identified_resources([(uri, resource)])

    def with_identified_resources(
        self,
        pairs: Iterable[tuple[str, IdentifiedResource]],
    ) -> Registry:
        uncrawled = self._uncrawled.evolver()
        contents = self._contents.evolver()
        for uri, resource in pairs:
            if uri not in self._contents or self._contents[uri][0] != resource:
                uncrawled.add(uri)
            contents[uri] = resource, m()
        return evolve(
            self,
            contents=contents.persistent(),
            uncrawled=uncrawled.persistent(),
        )

    def resource_at(
        self,
        uri: str,
    ) -> tuple[IdentifiedResource, PMap[str, AnchorType], Registry]:
        at_uri = self._contents.get(uri)
        if at_uri is None or uri in self._uncrawled:
            registry = self._crawl()
            return *registry._contents[uri], registry
        return *at_uri, self

    def _crawl(self) -> Registry:
        contents = self._contents.evolver()
        resources = [(uri, contents[uri][0]) for uri in self._uncrawled]
        while resources:
            uri, resource = resources.pop()
            anchors = pmap((each.name, each) for each in resource.anchors())

            id = resource.id()
            if id is None:
                if anchors:
                    old, old_anchors = contents[uri]
                    contents[uri] = old, old_anchors.update(anchors)
            else:
                uri = urljoin(uri, id)
                contents[uri] = resource, anchors

            resources.extend((uri, each) for each in resource.subresources())
        return evolve(self, contents=contents.persistent(), uncrawled=s())

    def resolver(self, root: Schema, specification: Specification) -> Resolver:
        uri = specification.id_of(root) or ""
        registry = self.with_identified_resource(
            uri=uri,
            resource=IdentifiedResource(
                resource=root,
                specification=specification,
            ),
        )
        return Resolver(base_uri=uri, registry=registry)


@define
class Resolver:

    _base_uri: str
    _registry: Registry
    _previous: PList[str] = field(default=plist(), repr=False)

    def lookup(self, ref: str) -> tuple[Schema, Resolver]:
        if ref.startswith("#"):
            uri, fragment = self._base_uri, ref[1:]
        else:
            uri, fragment = urldefrag(urljoin(self._base_uri, ref))

        resource, anchors, registry = self._registry.resource_at(uri)

        if fragment.startswith("/"):
            base_uri, target = resource.pointer(base_uri=uri, pointer=fragment)
        else:
            if fragment:
                resource, uri = anchors[fragment].resolve(
                    resolver=self,
                    uri=uri,
                )

            base_uri, target = uri, resource.resource
            id = resource.id()
            if id is not None:
                base_uri = urljoin(self._base_uri, id).rstrip("#")
        return target, self._evolve(base_uri=base_uri, registry=registry)

    def with_root(
        self,
        root: Schema,
        specification: Specification,
    ) -> Resolver:
        maybe_relative = specification.id_of(root)
        if maybe_relative is None:
            return self

        uri = urljoin(self._base_uri, maybe_relative)
        registry = self._registry.with_identified_resource(
            uri=uri,
            resource=IdentifiedResource(
                resource=root,
                specification=specification,
            ),
        )
        return self._evolve(base_uri=uri, registry=registry)

    def dynamic_scope(
        self,
    ) -> Iterable[tuple[str, IdentifiedResource, PMap[str, AnchorType]]]:
        for uri in self._previous:
            resource, anchors, _ = self._registry.resource_at(uri)
            yield uri, resource, anchors

    def _evolve(self, **kwargs):
        previous = self._previous.cons(self._base_uri)
        return evolve(self, previous=previous, **kwargs)
